/*
 * Copyright (c) 2007 Koji Suzuki
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <string.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/eeprom.h>
#include <avr/wdt.h>
#include <util/delay.h>

#include "usbconfig.h"
#include "cdc162.h"

#define bit_set(port,bit)	{ port |= (1<<(bit));}
#define bit_clear(port,bit)	{ port &= ~(1<<(bit));}
#define min(a,b)		((a)<(b) ? (a):(b))


#define nop()	__asm__ volatile ("nop" )

#include "serflash.h"

void init_spi2() {
	bit_clear(SPI2_DDR, SPI2_MISO);
	SPI2_PORT &=  ~((1<<SPI2_SCK) | (1<<SPI2_MOSI) |(1<<SPI2_MISO) );
	SPI2_DDR |=    ((1<<SPI2_SCK) | (1<<SPI2_MOSI));
	bit_set(SPI2_CS_PORT, SPI2_CS);
	bit_set(SPI2_CS_DDR, SPI2_CS);
#ifdef HARD_SPI2
	// CLK = 8MHz ( 16MHz / 2 )
	UBRR1 = 0;
	UCSR1C = (1<<UMSEL11)|(1<<UMSEL10);
	UCSR1B = (1<<RXEN1) | (1<<TXEN1);
#endif
}

void term_spi2() {
#ifdef HARD_SPI2
	UCSR1B = 0;
	UCSR1C = 0;
#endif
	SPI2_DDR &=   ~((1<<SPI2_SCK) | (1<<SPI2_MOSI) | (1<<SPI2_MISO));
	SPI2_CS_DDR &=   ~(1<<SPI2_CS);
	SPI2_PORT &=  ~((1<<SPI2_SCK) | (1<<SPI2_MOSI) | (1<<SPI2_MISO));
	bit_clear(SPI2_CS_DDR, SPI2_CS);
	bit_clear(SPI2_CS_PORT, SPI2_CS);
}

#ifdef HARD_SPI2
uint8_t spi2_write(uint8_t c) {
	 UDR1 = c;
	 while (bit_is_clear(UCSR1A, RXC1))
		 ;
	 return UDR1;
}

#else
uint8_t spi2_write(uint8_t c) {
	uint8_t i,rx;
	i = 0x80;
	rx = 0;
	do {
		if (c & i) {
			bit_set(SPI2_PORT, SPI2_MOSI);
		} else {
			bit_clear(SPI2_PORT, SPI2_MOSI);
		}
		bit_set(SPI2_PORT, SPI2_SCK);
		if (bit_is_set(SPI2_PIN, SPI2_MISO)) {
			rx |= i;
		}
		bit_clear(SPI2_PORT, SPI2_SCK);
		i >>= 1;
	} while (i);
	return rx;
}
#endif

uint8_t serflash_signature[3];
uint8_t serflash_type;
uint16_t serflash_size; // 512B 

int8_t serflash_init() {
	uint8_t i;
	spi2_enable();
	spi2_write(SERFLASH_RDID);
	serflash_signature[0] = spi2_read();
	serflash_signature[1] = spi2_read();
	serflash_signature[2] = spi2_read();
	spi2_disable();
	if (serflash_signature[0] == 0x37) {
		if (serflash_signature[1] == 0x30) {
			if (serflash_signature[2] == 0x14) {
				serflash_type = SERFLASH_TYPE_A25L080;
			}
		}
	} else if (serflash_signature[0] == 0x1f) {
		if (serflash_signature[1] == 0x44) {
				serflash_type = SERFLASH_TYPE_AT25DF041;
		} else if (serflash_signature[1] == 0x45) {
				serflash_type = SERFLASH_TYPE_AT25DF081;
		} else if (serflash_signature[1] == 0x46) {
				serflash_type = SERFLASH_TYPE_AT25DF161;
		} else if (serflash_signature[1] == 0x47) {
				serflash_type = SERFLASH_TYPE_AT25DF321;
		}
#ifndef SUPPORT_ERASE4K
	} else if (serflash_signature[0] == 0x20) {
		if (serflash_signature[1] == 0x20) {
			if (serflash_signature[2] == 0x14) {
				serflash_type = SERFLASH_TYPE_M25P80;
			} else if (serflash_signature[2] == 0x13) {
				serflash_type = SERFLASH_TYPE_M25P40;
			} else if (serflash_signature[2] == 0x12) {
				serflash_type = SERFLASH_TYPE_M25P20;
			}
		}
	} else {
		spi2_enable();
		spi2_write(SERFLASH_RES);
		for (i=0; i<3; i++) {
			spi2_read();
		}
		i = spi2_read();
		spi2_disable();
		if (i == 0x11) {
			serflash_type = SERFLASH_TYPE_M25P20;
		} else if (i == 0x12) {
			serflash_type = SERFLASH_TYPE_M25P40;
		} else if (i == 0x13) {
			serflash_type = SERFLASH_TYPE_M25P80;
		}
#endif
	}
	if (serflash_type) {
		serflash_size = 512 << (serflash_type & 0xf);
	}
	return serflash_type;
}

void serflash_open(uint8_t cmd , int32_t offs) {
	spi2_enable();

	spi2_write(cmd);
	spi2_write((offs >> 16) & 0xff);
	spi2_write((offs >> 8) & 0xff);
	spi2_write(offs  & 0xff);
}
